/*
 * (C)2013 IBM Corporation
 */
#ifndef _REPO_RGLOBAL_H
#define _REPO_RGLOBAL_H


#include <string>
#include "ArgParser.h"


#if defined(R_ARCH) && (R_ARCH == 1)
#define R_HAS_FUSIONIO
#define R_HAS_MELLANOX
#ifdef R_PREBOOT
#define R_HAS_INTELNIC
#define R_HAS_NVIDIA
#endif
#endif


#define LIBREPO_BEGIN_NAMESPACE   namespace onecli { namespace repository {
#define LIBREPO_END_NAMESPACE     } }


#ifdef _WIN32
#ifdef LIBREPO_STATIC
#define LIBREPO_API
#else
#ifdef LIBREPO_EXPORTS
#define LIBREPO_API __declspec(dllexport)
#else
#define LIBREPO_API __declspec(dllimport)
#endif
#endif
#define LIBREPO_LOAD_SYMBOL  GetProcAddress
#define LIBREPO_FREE FreeLibrary
#else
#if __GNUC__ >= 4
#define LIBREPO_API __attribute__ ((visibility("default")))
#else
#define LIBREPO_API
#endif
#define LIBREPO_LOAD_SYMBOL  dlsym
#define LIBREPO_FREE dlclose
#endif


#ifdef _WIN32
#pragma warning(disable: 4996)
#ifndef _DLL /* msvc assumed */
#error CRT dynamic link is required.
#endif
#endif


LIBREPO_BEGIN_NAMESPACE


/** String pair */
typedef std::pair<std::string, std::string> RStringPair;
/** Callback for running inventory */
typedef void (*RInventoryCallback)(const std::string &type, const std::string &display);

#if 0
/**
 * system type
 */
typedef enum _RSystemType {
    UnKnownSYS = 0, // this is initial value
    SystemX, //target system is SYSTEM X. normally, it is default value
    ThinkServer  //target system is ThinkServer system
} RSystemType;
#endif

/**
 * Connection mode
 */
typedef enum _RConnectionMode {
    IBMode = 0,
    OOBMode,  /* connect to imm cim server under windows/linux. */
    ESXiMode,
    AgentlessMode,
    CMMMode
} RConnectionMode;

/*for each case, inventory has different XML schema for it, which it  means different XML report, HTML report*/
enum RInventoryCase {
    IB_WINDOWS = 0,
    IB_SLES,
    IB_RHEL,
    IB_PREBOOT,
    OOB_CASE,  
    ESXI_CASE,
    CMM_CASE,
    THINK_SERVER_WIN,
    THINK_SERVER_SLES,
    THINK_SERVER_RHEL,
    INVENTORY_CASE_MAX
};

/**
 * Connection type
 */
typedef enum _RConnectionType {
    HTTP,
    HTTPS
} RConnectionType;

/**
 * Connection Info
 */
typedef struct _RConnectionInfo {
    RConnectionMode mode;
    RConnectionType type;
    std::string ipaddr;
    unsigned short port;
    std::string user;
    std::string pass;
    CLParamIPVersion ipversion;
    _RConnectionInfo()
            : mode(IBMode), type(HTTPS), ipaddr(""), port(0), user(""), pass(""), ipversion(PARAM_IPV4) {
    }
    _RConnectionInfo(RConnectionMode _mode, RConnectionType _type,
                     const std::string &_ipaddr, unsigned short _port, const std::string &_user, const std::string &_pass, CLParamIPVersion _ipversion)
            : mode(_mode), type(_type), ipaddr(_ipaddr), port(_port), user(_user), pass(_pass), ipversion(_ipversion) {
    }
} RConnectionInfo;


LIBREPO_END_NAMESPACE


#endif
