## script to download package from sftp
ONECLI_NAME="onecli_for_bmu.tgz"
D_ONECLI="onecli"
D_PACKAGE="packages"
OUTPUT_COMPARE="output_compare"
OUTPUT_FLASH="output_flash"
USER="user"
PASS="pass"
SFTP_IP="10.240.194.21"
SFTP_PORT="6990"
SFTP_CWD="/root/"
SUBCOMMAND="compare"
SELECTED_PACKAGES=""
ALL_XMLS=""
set -x

function yell(){
    echo "$0: $*" >&2
}
function die(){
    yell "$*"; exit 1
}
function try(){
    "$@" || die "cannot $*"
}

function make_work_dirs(){
    mkdir $D_ONECLI
    mkdir $D_PACKAGE
    mkdir $OUTPUT_FLASH
    mkdir $OUTPUT_COMPARE
    ls >> $OUTPUT_FLASH/make_work_dirs.log
}

function download_onecli(){
    ## scp need user input password
    ##scp $USER:$PASS@$SFTP_IP:$SFTP_CWD/$ONECLI_NAME $D_ONECLI
    try curl -u $USER:$PASS sftp://$SFTP_IP:$SFTP_PORT$SFTP_CWD/$ONECLI_NAME --output $D_ONECLI/$ONECLI_NAME -k
    try tar xf $D_ONECLI/$ONECLI_NAME -C $D_ONECLI
    if [ ! -e $D_ONECLI/$ONECLI_NAME ];then
        echo "download $ONECLI_NAME failed" >> $OUTPUT_FLASH/download_onecli.log
    fi
}

function download_single_package(){
    ## $1 package or xml name
    #scp $USER:$PASS@$SFTP_IP:$SFTP_CWD/$1 $D_PACKAGE
    try curl -u $USER:$PASS sftp://$SFTP_IP:$SFTP_PORT$SFTP_CWD/$1 --output $D_PACKAGE/$1 -k
    if [ ! -e $D_PACKAGE/$1 ];then
        echo "download $1 failed" >> $OUTPUT_FLASH/download_single_package.log
    fi

}

function run_onecli_compare(){
    ## ibm test machine, need remove
    #echo "" > ./$D_ONECLI/IBM_systems_list.txt
    if [ "x$1" = "xforce" ];then
        try ./$D_ONECLI/OneCli update compare --lxca 1 --dir $D_PACKAGE --output $OUTPUT_COMPARE --log 5 --forceid all --scope latest
    else
        try ./$D_ONECLI/OneCli update compare --lxca 1 --dir $D_PACKAGE --output $OUTPUT_COMPARE --log 5 --scope latest
    fi
}

function download_selected_packages(){
    for package in $SELECTED_PACKAGES
    do
        download_single_package "$package"
    done
}

function flash_packages(){
    try ./$D_ONECLI/OneCli update flash --lxca 1 --dir $D_PACKAGE --comparexml $OUTPUT_COMPARE/Onecli-update-compare.xml --output $OUTPUT_FLASH/ --log 5
	rm -rf $OUTPUT_COMPARE/Onecli-update-compare.xml 
}

function flash_by_id(){
    echo $1
    try ./$D_ONECLI/OneCli update flash --lxca 1 --dir $D_PACKAGE --output $OUTPUT_FLASH/ --log 5 --nocompare --includeid $1
}

LOG_ZIP="bmu_log.tar.gz"
function zip_output(){
    cp -rf ./$OUTPUT_COMPARE ./$OUTPUT_FLASH/
    tar zcvf $LOG_ZIP $OUTPUT_FLASH > /dev/null
}

function query_usb0_ip(){
    (ipmitool raw 0x0C 0x02 0x09 0xCB 0x00 0x00 | tee /var/log/imm-ip.log) >> $OUTPUT_FLASH/query_usb0_ip.log
    imm_hex_ip=(`(cat /var/log/imm-ip.log | cut -d ' ' -f 3-6) >> $OUTPUT_FLASH/query_usb0_ip.log`)
    length=${#imm_hex_ip[@]}
    imm_ip="169.254.95.118"
    if [ $length = 4 ];then
        imm_ip="`printf %d 0x${imm_hex_ip[0]}`.`printf %d 0x${imm_hex_ip[1]}`.`printf %d 0x${imm_hex_ip[2]}`.`printf %d 0x${imm_hex_ip[3]}`"
    fi
    SFTP_IP=$imm_ip
    echo "USB0_IP=$SFTP_IP" >> $OUTPUT_FLASH/query_usb0_ip.log
}

function run_cmd(){
    ### $@ = subcommand [options]
    SUBCOMMAND=$1
    shift 1
    case $SUBCOMMAND in
    "make_work_dirs")
        make_work_dirs
        ;;
    "download_onecli")
    ### $@ = sftp_usr sftp_pass sftp_pwd [Onecli package name]
        USER=$1
        PASS=$2
        SFTP_CWD=$3
        query_usb0_ip
        if [ ! -n "$SFTP_IP" ]; then
            ## if query ip faild user default ip
            SFTP_IP="169.254.95.118"
        fi
        shift 3
        if [[ $# > 0 ]]
        then
            ONECLI_NAME=$1
        fi
        download_onecli
        ;;
    "compare")
        run_onecli_compare $@
        ;;
     "download_selected")
     ### $@ = sftp_usr sftp_pass sftp_pwd [packages]
        USER=$1
        PASS=$2
        SFTP_CWD=$3
        query_usb0_ip
        if [ ! -n "$SFTP_IP" ]; then
            ## if query ip faild user default ip
            SFTP_IP="169.254.95.118"
        fi
        shift 3
        for package in "$@"
        do
            SELECTED_PACKAGES="$SELECTED_PACKAGES $package"
            ALL_XMLS="$ALL_XMLS $package"
        done
        download_selected_packages
        ;;
     "flash")
         flash_packages
         ;;
     "zip_output")
     ### $@ = output_zip_name
         LOG_ZIP=$1
         zip_output
         ;;
    "flash_by_id")
         flash_by_id $@
         ;;
     "test")
        query_usb0_ip
        echo $SFTP_IP
    esac
}

run_cmd "$@"




