// Copyright 2016 Lenovo
;(function() {
'use strict';

// Define all fields
var fields = {
    groupname: {
        defvalue: null,
        label: 'Group Name',
        apiname: 'name',
        required: true,
        readonly: false,
        type: 'text',
        placeholder: null
    },
    BMC: {
        defvalue: null,
        label: 'BMC',
        apiname: 'hardwaremanagement.manager',
        required: false,
        readonly: false,
        type: 'text',
        placeholder: null
    },
    BMCuser: {
        defvalue: null,
        isset: false,
        label: 'BMC Username',
        apiname: 'secret.hardwaremanagementuser',
        required: false,
        readonly: false,
        type: 'text',
        placeholder: null
    },
    BMCpassword: {
        defvalue: null,
        isset: false,
        label: 'BMC Password',
        apiname: 'secret.hardwaremanagementpassword',
        required: false,
        readonly: false,
        type: 'password',
        placeholder: null
    },
    rack: {
        defvalue: null,
        label: 'Rack',
        apiname: 'location.rack',
        required: false,
        readonly: false,
        type: 'text',
        placeholder: null
    },
    u: {
        defvalue: null,
        label: 'U',
        apiname: 'location.u',
        required: false,
        readonly: false,
        type: 'text',
        placeholder: null
    },
    enclosure: {
        defvalue: null,
        label: 'Enclosure',
        apiname: 'enclosure.manager',
        required: false,
        readonly: false,
        type: 'text',
        placeholder: null
    },
    bay: {
        defvalue: null,
        label: 'Enclosure Bay',
        apiname: 'enclosure.bay',
        required: false,
        readonly: false,
        type: 'text',
        placeholder: null
    },
    model: {
        defvalue: null,
        label: 'Model',
        apiname: 'id.model',
        required: false,
        readonly: false,
        type: 'text',
        placeholder: null
    },
    nodes: {
        defvalue: null,
        label: 'Nodes',
        apiname: 'nodes',
        required: false,
        readonly: false,
        type: 'text',
        placeholder: 'n1, n2, n3'
    }

};

function addGroup() {
    var uidlg = this;
    var inputs = uidlg.getElementsByTagName('input');
    //var jsontext = '{';
    var json = {};
    for (var i=0; i < inputs.length; i++) {
        var field = inputs[i].id;
        var apifield = fields[field].apiname;
        var apival = inputs[i].value;
        var required = inputs[i].required;

        if(apival == null){  // send empty strings, not null values
            apival = "";
        }

        if(apival == "" && required){   // required field is empty, return without saving
            alert("Please enter a value for all required (*) fields.");
            return false;
        }


        if(apifield != null && apifield != "null" && apival != ""){   // create json to send to api and update fields json

            if(field == "nodes"){   // parse through list, convert to array
                var nodesNoSpaces = apival.replace(/ /g, "");   // remove all spaces, so that "n1, n2, n3" and "n1,n2,n3" are treated the same
                var nodesArray = nodesNoSpaces.split(",");  // split into an array
                apival = nodesArray;
            }

            json[apifield] = apival;
        }

        // update fields
        fields[field].defvalue = apival;

    }

    postRequest('/confluent-api/nodegroups/', json, refresh);

}


function clearGroupUI(){
    for (var field in fields){
        fields[field].defvalue = null;
    }
}

function addGroupUI() {
    var uidlg = document.createElement('div');      // create dialog div
    uidlg.classList.add('modaldialog');

    var heading = uidlg.appendChild(document.createElement("span"));
    heading.classList.add("noteHeader");
    heading.innerHTML = "Add Group";

    var dialogContent = document.createElement('div');

    // add content div
    uidlg.appendChild(dialogContent);

    // fill content
    fillDataFromJson(dialogContent, true);

    var bg = document.createElement('div');
    bg.classList.add('modalbg');
    document.body.appendChild(bg);
    bg.appendChild(uidlg);
    uidlg.appendChild(document.createElement('hr'));
    var save = document.createElement('button');
    save.textContent = l('Save');
    save.addEventListener('click', closeDialog.bind([uidlg, bg, addGroup.bind(uidlg)]));
    uidlg.appendChild(save);
    var cancelbtn = document.createElement('button');
    cancelbtn.textContent = l('Cancel');
    cancelbtn.addEventListener('click', closeDialog.bind([uidlg, bg, clearGroupUI]));
    uidlg.appendChild(cancelbtn);
}


function fillDataFromJson(dialog, initial){


    if(initial == null){
        initial = false;
    }

    if(dialog == null){
        dialog = this;
    }

    dialog.innerHTML = "";

    // loop through all fields, defined above. Load info (label and input) into dialog box
    for (var field in fields) {
        var curval = localStorage.getItem(field);   // I don't think this is applicable for new groups since we won't be saving
        if (curval === null) {
            curval = fields[field].defvalue;
        }


        // set up initial requirements
        if(initial == true){

            fields['groupname'].required = true;

        }

        // set up all inputs and labels
        var curinput = document.createElement('input');
        //curinput.id = fields[field].apiname;
        curinput.id = field;
        if (curval !== null) {
            curinput.value = curval;
        }
        curinput.required = fields[field].required;
        curinput.readOnly = fields[field].readonly;
        curinput.type = fields[field].type;
        if(fields[field].placeholder != null) curinput.placeholder = fields[field].placeholder;
        var thelab = document.createElement('label');
        thelab.setAttribute('for', curinput.id);
        thelab.textContent = l(fields[field].label) + ':';

        if(curinput.required == true){
            thelab.textContent = "* " + thelab.textContent;
        }

        dialog.appendChild(thelab);
        dialog.appendChild(curinput);
        dialog.appendChild(document.createElement('br'));
    }

}


var refreshTimeout = 0;

function refresh(){

    clearGroupUI();

    var hash = window.location.hash;

    if(hash == "#compactoverview" || hash == ""){

        if(refreshTimeout == 0){
            refreshTimeout = setTimeout(function(){window.location.reload(true)}, 1);
        }
    }
    
}

window.addGroupUI = addGroupUI;
})();

