// Copyright 2016 Lenovo
;(function() {
'use strict';

// Define all fields
var fields = {
    nodename: {
        defvalue: null,
        label: 'Node Name',
        apiname: 'name',
        required: true,
        readonly: false,
        type: 'text'
    },
    BMC: {
        defvalue: null,
        label: 'BMC',
        apiname: 'hardwaremanagement.manager',
        required: true,
        readonly: false,
        type: 'text'
    },
    BMCuser: {
        defvalue: null,
        isset: false,
        label: 'BMC Username',
        apiname: 'secret.hardwaremanagementuser',
        required: true,
        readonly: false,
        type: 'text'
    },
    BMCpassword: {
        defvalue: null,
        isset: false,
        label: 'BMC Password',
        apiname: 'secret.hardwaremanagementpassword',
        required: true,
        readonly: false,
        type: 'password'
    },
    rack: {
        defvalue: null,
        label: 'Rack',
        apiname: 'location.rack',
        required: false,
        readonly: false,
        type: 'text'
    },
    u: {
        defvalue: null,
        label: 'U',
        apiname: 'location.u',
        required: false,
        readonly: false,
        type: 'text'
    },
    enclosure: {
        defvalue: null,
        label: 'Enclosure',
        apiname: 'enclosure.manager',
        required: false,
        readonly: false,
        type: 'text'
    },
    bay: {
        defvalue: null,
        label: 'Enclosure Bay',
        apiname: 'enclosure.bay',
        required: false,
        readonly: false,
        type: 'text'
    },
    model: {
        defvalue: null,
        label: 'Model',
        apiname: null,
        required: false,
        readonly: false,
        type: 'text'
    },
    serial: {
        defvalue: null,
        label: 'Serial',
        apiname: null,
        required: false,
        readonly: false,
        type: 'text'
    },
    MAC: {
        defvalue: null,
        label: 'MAC',
        apiname: null,
        required: false,
        readonly: false,
        type: 'text'
    },
    groups: {
        defvalue: null,
        label: 'Nodegroups',
        apiname: 'groups',
        required: false,
        readonly: false,
        type: 'text',
        placeholder: 'group1, group2, group3'
    }

};

function addNode() {
    var uidlg = this;
    var inputs = uidlg.getElementsByTagName('input');
    //var jsontext = '{';
    var json = {};
    for (var i=0; i < inputs.length; i++) {
        var field = inputs[i].id;
        var apifield = fields[field].apiname;
        var apival = inputs[i].value;
        var required = inputs[i].required;

        if(apival == null){  // send empty strings, not null values
            apival = "";
        }

        if(apival == "" && required){   // required field is empty, return without saving
            alert("Please enter a value for all required (*) fields.");
            return false;
        }


        if(apifield != null && apifield != "null" && apival != ""){   // create json to send to api and update fields json
            
            if(apifield == "groups"){
                var groups = apival.replace(", ", ",").split(",");
                apival = groups;
            }

            json[apifield] = apival;
        }

        // update fields
        fields[field].defvalue = apival;

    }

    postRequest('/confluent-api/nodes/', json, assignNode);

}


var selects = new Array();
var labels = new Array();

function discover(){    // kicks off discovery. Called once.

    var callerDialog = this;

    changeRequirements(callerDialog, ["nodename", "MAC", "BMC", "BMCuser", "BMCpassword"]);

    // set up dialog
    var dialog = document.createElement('div');
    dialog.classList.add('modaldialog');

    var bg = document.createElement('div');
    bg.classList.add('modalbg');
    document.body.appendChild(bg);
    bg.appendChild(dialog);

    var heading = dialog.appendChild(document.createElement("span"));
    heading.classList.add("noteHeader");
    heading.innerHTML = "Discover";
    var headBreak = dialog.appendChild(document.createElement("span"));
    headBreak.innerHTML = "<br/>"; 

    var dialogContent = document.createElement('div');
    dialog.appendChild(dialogContent);

    dialog.appendChild(document.createElement('br'));
    
    dialog.appendChild(document.createElement('hr'));
    var save = document.createElement('button');
    save.id = "saveDiscover";
    save.textContent = l('OK');
    save.addEventListener('click', closeDialog.bind([dialog, bg, fillData.bind(callerDialog)]));
    save.disabled = true;
    dialog.appendChild(save);
    var cancelbtn = document.createElement('button');
    cancelbtn.textContent = l('Cancel');
    cancelbtn.addEventListener('click', closeDialog.bind([dialog, bg, clearDiscover]));
    dialog.appendChild(cancelbtn);

    // start with one drop down -- how to search?

    discoverNext(dialogContent);

}


function discoverNext(dialog){   // called for each new dropdown. 

    if(dialog == null)
        var dialog = this;

    // set up new select
    var index = selects.length;  // haven't added new one, so no need to -1

    var newSelect = document.createElement("select");
    newSelect.id = "select" + index;
    newSelect.addEventListener("change", processChange.bind([index, dialog]));

    var blankOpt = document.createElement("option");
    blankOpt.innerHTML = " ";
    blankOpt.value = "";
    newSelect.appendChild(blankOpt);
    
    var newLabel = document.createElement("label");
    //newLabel.innerHTML = "Select " + index;

    dialog.appendChild(newLabel);
    dialog.appendChild(newSelect);
    dialog.appendChild(document.createElement('br'));
    selects.push(newSelect);

    var url = createURL(index);

    // call new request
    getRequest(url, processDiscovery, failDiscovery);



    setTimeout(function(){if(!hitNode)editNewDropdown(newLabel, index, dialog);}, 500);

}

function editNewDropdown(label, index, dialog){
    label.innerHTML = labels[index];

    if(selects[index].options.length == 2){    // only one actual option - autofill

        selects[index].options.selectedIndex = 1;

        processChange(index, dialog);

    }else if(selects[index].options.length == 1){   // no options (only blank) -- remove dropdown and display error message


        var lastSelect = selects.pop();   // remove last select from select array
        var lastLabel = labels.pop();

        // remove label and select from dialog
        // last 3 nodes are label, dialog, br.
        var parentNode = lastSelect.parentNode;
        parentNode.removeChild(parentNode.lastChild);  // remove br
        parentNode.removeChild(parentNode.lastChild);  // remove select
        parentNode.removeChild(parentNode.lastChild);  // remove label

        var error = parentNode.appendChild(document.createElement("span"));
        error.innerHTML = "No values. Try again.";


    }
}

function getLabel(text, index){

    var label = "";

    var substring = "";

    if(text.length >= 2){
        substring = text.substring(0,2);

        if(substring == "by"){   // by mac, by serial, etc. Label should be "filter by: "

            label = "Filter by: ";

        }else{      // the MAC, serial, etc itself. Find out what it is based on previous select

            var selectedIndex = selects[index-1].options.selectedIndex;

            label = selects[index-1].options[selectedIndex].innerHTML + ": ";

        }
    }

    return label;
}

function processChange(index, dialog){

    if (index == null || index instanceof Event)
        index = this[0];
    if (dialog == null)
        dialog = this[1];

    // remove "no values" error
    if(dialog.lastChild.nodeName == "SPAN"){
        dialog.removeChild(dialog.lastChild);
    }

    for(var i=selects.length-1; i>index; i--){
        var lastSelect = selects.pop();   // remove last select from select array
        var lastLabel = labels.pop();

        // remove label and select from dialog
        // last 3 nodes are label, dialog, br.
        var parentNode = lastSelect.parentNode;
        parentNode.removeChild(parentNode.lastChild);  // remove br
        parentNode.removeChild(parentNode.lastChild);  // remove select
        parentNode.removeChild(parentNode.lastChild);  // remove label

    }

    if(hitNode){   // if we already hit a node but now we're changing upstream, want to reset
        hitNode = false;  
        nodeInfo = {};

    }

    
    discoverNext(dialog);

}

function changeRequirements(dialog, requiredFields){

    // loop through all fields, set required to false
    for (var field in fields) {
        
        // if (field == 'nodename'){
        //     fields[field].required = true;
        // }else{
        //     fields[field].required = false;
        // }

        if (requiredFields.includes(field)){
            fields[field].required = true;
        }else{
            fields[field].required = false;
        }
        

    }

    // remove * from all labels

    // var labs = dialog.getElementsByTagName('label')

    // for (var i = 0; i < labs.length; i++){

    //     var lab = labs[i];
    //     var newLabel = lab.innerHTML.replace('* ', '');
    //     lab.innerHTML = newLabel;

    // }


}

function createURL(index){

    var url = "/confluent-api/discovery/";   // base url

    for(var i=0; i<=index; i++){
        url += selects[i].value;              // add url from each dropdown's value
    }


    return url;

}

var hitNode = false; 

function processDiscovery(data){    // called for each new option in the dropdown


    for(var item in data){

        if(item == '_requestdone'){
            return;
        }

        if(data[item].href != null){   // keep going, not at node yet

            var href = data[item].href;

            if(href != "rescan"){       // skip rescan

                var newOpt = document.createElement('option');
                newOpt.id = href;
                newOpt.value = href;
                // newOpt.innerHTML = href.replace("-", " ").replace("/", "").replace("by ", "").toUpperCase();
                newOpt.innerHTML = href.replace("/", "").replace("by-", "").toUpperCase();

                //document.getElementById("select0").appendChild(newOpt);
                var index = selects.length - 1;
                var select = selects[index];  // get latest select
                select.appendChild(newOpt);

                var label = getLabel(href, index);
                if(labels[index] == "" || labels[index] == null){
                    labels[index] = label;
                }

            }



        }else{   // hit the node

            if(hitNode != true){   // this is the first hit

                // remove the new select
                var lastSelect = selects.pop();   // remove last select from select array
                var lastLabel = labels.pop();

                // remove label and select from dialog
                // last 3 nodes are label, dialog, br.
                var dialog = lastSelect.parentNode;
                dialog.removeChild(dialog.lastChild);  // remove br
                dialog.removeChild(dialog.lastChild);  // remove select
                dialog.removeChild(dialog.lastChild);  // remove label

                // enable "ok" button
                document.getElementById("saveDiscover").disabled = false;


                hitNode = true;  // will only do this once

            }

            
            compileNode(data);


        }
        
    }

}

var nodeInfo = new Object();

function compileNode(data){

    for (var item in data){

        if(item == '_requestdone'){
            return;
        }

        var property = item;
        var value = data[item];

        nodeInfo[property] = value;

    }

}

function failDiscovery(){

    console.log("Discovery failed.");

}


function fillData(){

    var callerDialog = this;

    var node = nodeInfo["nodename"];

    // get discovery data
    getDiscoveryData(callerDialog);


    // get node data if node was defined, otherwise just fill in data from discovery
    if(node == null || node == ""){   // can't get any more node info

        fields["nodename"].required = true;   // require user enter nodename

        fillDataFromJson(callerDialog, false);

    }else{

        fields["nodename"].defvalue = node;

        // for BMC info, not required if already set. But still changeable (not readonly)

        if(fields["BMC"].isset == true){
            fields["BMC"].required = false;
        }
        if(fields["BMCuser"].isset == true){
            fields["BMCuser"].required = false;
        }
        if(fields["BMCpassword"].isset == true){
            fields["BMCpassword"].required = false;
        }

        getNodeData(node, callerDialog);
    }


    clearDiscover();

}


function assignNode(result){

    if(! ('_requestdone' in result)){
        return;
    }

    var mac = fields["MAC"].defvalue.replace(/:/g, '-');
    var nodename = fields["nodename"].defvalue;


    if(mac != null && mac != ""){

        var url = '/confluent-api/discovery/by-mac/' + mac;
        var json = {'node': nodename};

        postRequest(url, json, refresh);

    }else{    // no mac to assign

        console.log("no mac");

        refresh();

    }

}



function getDiscoveryData(dialog){

    fields["MAC"].defvalue = nodeInfo["macs"][0];
    fields["model"].defvalue = nodeInfo["modelnumber"];
    fields["serial"].defvalue = nodeInfo["serialnumber"];

    fields["MAC"].readonly = true;

}

function clearDiscover(){
    nodeInfo = {};
    hitNode = false;
    selects = [];
    labels = [];
}

function clearNodeUI(){
    for (var field in fields){
        fields[field].defvalue = null;
    }
}

function addNodeUI() {
    var uidlg = document.createElement('div');      // create dialog div
    uidlg.classList.add('modaldialog');

    var heading = uidlg.appendChild(document.createElement("span"));
    heading.classList.add("noteHeader");
    heading.innerHTML = "Add Node";
    // var headBreak = uidlg.appendChild(document.createElement("span"));
    // headBreak.innerHTML = "<br/>"; 

    var dialogContent = document.createElement('div');

    // add discovery button to top
    var discoveryButton = document.createElement('button');
    discoveryButton.id = "discoveryButton";
    discoveryButton.innerHTML = "Discover";
    discoveryButton.addEventListener("click", discover.bind(dialogContent));
    uidlg.appendChild(discoveryButton);
    uidlg.appendChild(document.createElement('br'));

    // add content div
    uidlg.appendChild(dialogContent);

    // fill content
    fillDataFromJson(dialogContent, true);

    // loop through all fields, defined above. Load info (label and input) into dialog box
    // for (var field in fields) {
    //     var curval = localStorage.getItem(field);   // I don't think this is applicable for new nodes since we won't be saving
    //     if (curval === null) {
    //         curval = fields[field].defvalue;
    //     }

    //     // set up all inputs and labels
    //     var curinput = document.createElement('input');
    //     curinput.id = fields[field].apiname;
    //     if (curval !== null) {
    //         curinput.value = curval;
    //     }
    //     curinput.required = fields[field].required;
    //     var thelab = document.createElement('label');
    //     thelab.setAttribute('for', curinput.id);
    //     thelab.textContent = l(fields[field].label) + ':';
    //     uidlg.appendChild(thelab);
    //     uidlg.appendChild(curinput);
    //     uidlg.appendChild(document.createElement('br'));
    // }

    var bg = document.createElement('div');
    bg.classList.add('modalbg');
    document.body.appendChild(bg);
    bg.appendChild(uidlg);
    uidlg.appendChild(document.createElement('hr'));
    var save = document.createElement('button');
    save.textContent = l('Save');
    save.addEventListener('click', closeDialog.bind([uidlg, bg, addNode.bind(uidlg)]));
    uidlg.appendChild(save);
    var cancelbtn = document.createElement('button');
    cancelbtn.textContent = l('Cancel');
    cancelbtn.addEventListener('click', closeDialog.bind([uidlg, bg, clearNodeUI]));
    uidlg.appendChild(cancelbtn);
}


function fillDataFromJson(dialog, initial){


    if(initial == null){
        initial = false;
    }

    if(dialog == null){
        dialog = this;
    }

    dialog.innerHTML = "";

    // loop through all fields, defined above. Load info (label and input) into dialog box
    for (var field in fields) {
        var curval = localStorage.getItem(field);   // I don't think this is applicable for new nodes since we won't be saving
        if (curval === null) {
            curval = fields[field].defvalue;
        }


        // set up initial requirements
        if(initial == true){

            fields['nodename'].required = true;
            fields['BMC'].required = true;
            fields['BMCuser'].required = true;
            fields['BMCpassword'].required = true;


        }

        // set up all inputs and labels
        var curinput = document.createElement('input');
        //curinput.id = fields[field].apiname;
        curinput.id = field;
        if (curval !== null) {
            curinput.value = curval;
        }
        curinput.required = fields[field].required;
        curinput.readOnly = fields[field].readonly;
        curinput.type = fields[field].type;
        if(fields[field]["placeholder"] != null) curinput.placeholder = fields[field]["placeholder"];
        var thelab = document.createElement('label');
        thelab.setAttribute('for', curinput.id);
        thelab.textContent = l(fields[field].label) + ':';

        if(curinput.required == true){
            thelab.textContent = "* " + thelab.textContent;
        }

        dialog.appendChild(thelab);
        dialog.appendChild(curinput);
        dialog.appendChild(document.createElement('br'));
    }

}



function getNodeData(node, dialog){

    getRequest("/confluent-api/nodes/" + node + "/attributes/all", parseNodeData.bind(dialog));

}

function parseNodeData(json){

    var dialog = this;

    if ("_requestdone" in json){
        setTimeout(function(){fillDataFromJson(dialog);}, 100);
    }

    for (var property in json){

        var value = json[property]["value"];
        var isset = json[property]["isset"];

        switch(property){
            case "hardwaremanagement.manager":
                fields["BMC"].defvalue = value;
                break;
            case "secret.hardwaremanagementuser":
                fields["BMCuser"].isset = isset;
                break;
            case "secret.hardwaremanagementpassword":
                fields["BMCpassword"].isset = isset;
                break;
            case "location.rack":
                fields["rack"].defvalue = value;
                break;
            case "location.u":
                fields["u"].defvalue = value;
                break;
            case "enclosure.manager":
                fields["enclosure"].defvalue = value;
                break;
            case "enclosure.bay":
                fields["bay"].defvalue = value;
                break;
            default:
                break;
        }

  
    }

}



 //    enclosure: {
 //        defvalue: null,
 //        label: 'Enclosure',
 //        apiname: 'enclosure.manager',
 //        required: false
 //    },
 //    bay: {
 //        defvalue: null,
 //        label: 'Enclosure Bay',
 //        apiname: 'enclosure.bay',
 //        required: false
 //    }


var refreshTimeout = 0;

function refresh(){

    clearNodeUI();

    var hash = window.location.hash;

    if(hash == "#compactoverview" || hash == ""){

        if(refreshTimeout == 0){
            refreshTimeout = setTimeout(function(){window.location.reload(true)}, 1);
        }
    }
    
}

window.addNodeUI = addNodeUI;
})();

