;(function(){
    var nodelist;
    var cputempchart;
    var powerchart;
    var active = true;
    var currpower = [];
    var currpowernames = [];
    var currtemps = [];
    var currtempnames = [];
    var currfans = [];
    var currfannames = [];
    var mintemp = null;
    var maxtemp = null;
    var minpower = null;
    var maxpower = null;
    var minfan = null;
    var maxfan = null;
    var pendingpowerupdate = null;
    var pendingtempupdate = null;
    var pendingfanupdate = null;

    function pad(number) {
        if (number < 10) {
            return '0' + number;
        }
        return number;
    }

    function initpowerchart(pane) {
        nodelist = [];
        var node = getNodeName();  // is this sufficient?
        cputempchart = new_chart();
        powerchart = new_chart();
        pane.appendChild(cputempchart);
        pane.appendChild(powerchart);

        getRequest('/confluent-api/noderange/' + node + '/nodes/', updatenodelist);
    }

    function updatenodelist(data) {
        if (data.item) {
            var newnode = data.item.href.slice(0, -1);
            nodelist.push(newnode);
        }
        if ('_requestdone' in data) {
            updatepowerchart();
            updatecputempchart();
            updatefanchart();
        }
    }

    function new_chart() {
        var thechart = document.createElement('div');
        thechart.style.display = 'inline-block';
        thechart.style.width = (0.49 * window.innerWidth) + 'px';
        thechart.style.height = (0.8 * window.innerHeight) + 'px';
        thechart.data = null;
        return thechart;
    }

    function updatepowerchart() {
        if (!active) {
            return;
        }
        currpower = [];
        currpowernames = [];
        getRequest('/confluent-api/nodes/' + getNodeName() + '/sensors/hardware/power/all', parsePower);
    }

    function updatecputempchart() {
        if (!active) {
            return;
        }
        // Clear data from last time
        currtempnames = [];
        currtemps = [];
        getRequest('/confluent-api/nodes/' + getNodeName() + '/sensors/hardware/temperature/all', parseTemp);
    }

    function updatefanchart(){
        if (!active) {
            return;
        }
        // Clear data from last time
        currfannames = [];
        currfans = [];
        getRequest('/confluent-api/nodes/' + getNodeName() + '/sensors/hardware/fans/all', parseFans);
    }

    function parsePower(data){

        if(data.sensors != null){
            var sensors = data.sensors;
            sensors.forEach(function(sensor){
                var sensorName = sensor.name;
                var sensorValue = sensor.value;
                if(sensorValue != null){
                    addPowerToChart(sensorName, sensorValue);
                }
            });
        }

        // Update plot
        if ('_requestdone' in data){
            updateplot(powerchart, 'Node Power Consumption', 'Power (W)', minpower, maxpower, currpower, currpowernames, nodelist);
            pendingpowerupdate = setTimeout(updatepowerchart, 5000);
        }
    }

    function parseTemp(data){

        if(data.sensors != null){
            var sensors = data.sensors;
            sensors.forEach(function(sensor){
                var sensorName = sensor.name;
                var sensorValue = sensor.value;
                if(sensorValue != null){
                    addTempToChart(sensorName, sensorValue);
                }
            });
        }

        // Update plot
        if ('_requestdone' in data){
            updateplot(cputempchart, 'Temperature and Fan Speed', 'Degrees (C)', mintemp, maxtemp, currtemps, currtempnames, nodelist);
            pendingtempupdate = setTimeout(updatecputempchart, 5000);
        }
    }

    function parseFans(data){
        if(data.sensors != null){
            var sensors = data.sensors;
            sensors.forEach(function(sensor){
                var sensorName = sensor.name;
                var sensorValue = sensor.value;
                if(sensorValue != null){
                    addFanToChart(sensorName, sensorValue);
                }
            });
        }

        // Update plot
        if ('_requestdone' in data){
            updateplot(cputempchart, 'Temperature and Fan Speed', 'RPM', minfan, maxfan, currfans, currfannames, nodelist);
            pendingfanupdate = setTimeout(updatefanchart, 5000);
        }
    }

    function addPowerToChart(currentName, currentValue){

        // Add current power & name to lists
        if(!currpowernames.includes(currentName)){
            currpowernames.push(currentName);
            currpower.push(currentValue);
        }

        // Update min and max power
        if (minpower == null){
            minpower = currentValue;
        } else if (currentValue < minpower){
            minpower = currentValue;
        }
        if (maxpower == null){
            maxpower = currentValue;
        } else if (currentValue > maxpower){
            maxpower = currentValue;
        }
    }

    function addTempToChart(currentName, currentValue){

        // Add current temp & name to lists
        if(!currtempnames.includes(currentName)){
            currtempnames.push(currentName);
            currtemps.push(currentValue);
        }        

        // Update min and max temps
        if (mintemp == null){
            mintemp = currentValue;
        } else if (currentValue < mintemp){
            mintemp = currentValue;
        }
        if (maxtemp == null){
            maxtemp = currentValue;
        } else if (currentValue > maxtemp){
            maxtemp = currentValue;
        }
    }

    function addFanToChart(currentName, currentValue){

        // Add current fan & name to lists
        if(!currfannames.includes(currentName)){
            currfannames.push(currentName);
            currfans.push(currentValue);
        }        

        // Update min and max fans
        if (minfan == null){
            minfan = currentValue;
        } else if (currentValue < minfan){
            minfan = currentValue;
        }
        if (maxfan == null){
            maxfan = currentValue;
        } else if (currentValue > maxfan){
            maxfan = currentValue;
        }
    }

    var metrics = [];    // json
    var allMetrics = []; // just to keep track of the order they were added
    
    function updateplot(plot, title, metric, min, max, data, tracenames, elemlist){
        var thedata;
        var layout;

        if (plot.data == null || metrics[title] == null || metrics[title][metric] == null){  // either nothing or haven't seen this metric yet

           // set up thedata
            thedata = [];

            if(plot.data != null){
                thedata = plot.data;
            }

            
            for (var i = 0; i < data.length; i++){ // i represents the sensor
                thedata.push({
                    y: [],                // values for this sensor
                    name: tracenames[i],  // name of sensor
                    showscale: true,
                    opacity: 0.99,
                    x: [],
                    type: 'scatter'
                });
            }

            // set up layout
            if(plot.layout == null){            // don't overwrite existing layout
                layout = {
                    title: title,
                    xaxis: {
                        title: 'Time',
                        titlefont: {
                            family: 'Arial, sans-serif',
                            size: 14,
                            color: 'black'
                        },
                        type: 'category'
                    },
                    yaxis: {
                        titlefont: {
                            family: 'Arial, sans-serif',
                            size: 14,
                            color: 'black'
                        },
                        type: 'linear'
                    },
                    showlegend: true,
                    legend: {
                        x: 1.2,
                        y: 1
                    }
                };
            }else{
                layout = plot.layout;
            }


        } else {
            thedata = plot.data;
            layout = plot.layout;
        }


        // Set min and max
        // thedata[0].cmin = min;
        // thedata[0].cmax = max;

        // Get time information
        var now = new Date();
        var currentTime = now.getHours() + ":" + pad(now.getMinutes()) + ":" + pad(now.getSeconds());

        // Add new information
        var nextData;
        

        for (var i = 0; i < data.length; i++){  // i represents the sensor
            nextData = data[i];  // the sensor's reading
        
            // get trace
            var traceIndex = -1;

            for(var j = 0; j < thedata.length; j++){
                if(thedata[j].name == tracenames[i]){
                    traceIndex = j;
                }
            }

            if(thedata[traceIndex].x.length >= 60){  // limit data points to 60
                thedata[traceIndex].y.shift();
                thedata[traceIndex].x.shift();
            }

            thedata[traceIndex].y.push(nextData);

            // use the same x axis for everything -- some data may be off by a second
            if(traceIndex == 0)
                thedata[traceIndex].x.push(currentTime);
            else
                thedata[traceIndex].x = thedata[0].x;

            // Set min and max
            // thedata[traceIndex].cmin = min;
            // thedata[traceIndex].cmax = max;


            // figure out y axis

            if(metrics[title] == null){              // check if already in metrics array and if not, add
                var newTitle = {};
                metrics[title] = newTitle;
            }
            if(metrics[title].hasOwnProperty(metric) == false){
                allMetrics.push(metric);
                metrics[title][metric] = allMetrics.indexOf(metric);  
            }
            
            if(metrics[title][metric] == minKey(metrics[title])){  // first (possibly only) axis

                if(thedata[i].yaxis == null){
                    thedata[i].yaxis = 'y';
                }
        
                layout.yaxis['title'] = metric;

                var scale = 0.2;
                var range = max-min;   
                    
                layout.yaxis.range = [min-scale*range, max+scale*range];


            }else{                                  // second axis -- assumes 2 axes max

                if(thedata[i].yaxis == null){
                    thedata[i].yaxis = 'y2';
                }


                if(layout.yaxis2 == null){

                    var yaxis2 = {};
                    layout["yaxis2"] = yaxis2;

                    layout.yaxis2 = {
                        title: metric,
                        overlaying: 'y',
                        side: 'right',
                        type: 'linear'
                    }

                }

                var scale = 0.2;
                var range = max-min;   

                layout.yaxis2.range = [min-scale*range, max+scale*range];
              
            }                
            
        }

        // Plot
        if (plot.data == null) {
            Plotly.newPlot(plot, thedata, layout);
        }else{
            Plotly.redraw(plot);
        }
    }

    function maxKey(obj){
        var max = 0;

        for(var i = 0; i < Object.keys(obj).length; i++){
            if(Object.keys(obj)[i] > max){
                max = Object.keys(obj)[i];
            }
        }

        return max;
    }

    function minKey(obj){
        var min = 100;
        var keys = Object.keys(obj);

        for(var i = 0; i < keys.length; i++){
            var key = keys[i];
            if(obj[key] < min){
                min = obj[key];
            }
        }

        return min;
    }

    window.initpowerchart = initpowerchart;

})();