// Copyright 2016 Lenovo
;(function() {
'use strict';

// Define all fields
var fields = {
    groupname: {
        defvalue: null,
        label: 'Group Name',
        apiname: 'name',
        required: true,
        readonly: false,
        type: 'select',
        placeholder: null
    },
    BMC: {
        defvalue: null,
        label: 'BMC',
        apiname: 'hardwaremanagement.manager',
        required: false,
        readonly: false,
        type: 'text',
        placeholder: null
    },
    BMCuser: {
        defvalue: null,
        isset: false,
        label: 'BMC Username',
        apiname: 'secret.hardwaremanagementuser',
        required: false,
        readonly: false,
        type: 'text',
        placeholder: null
    },
    BMCpassword: {
        defvalue: null,
        isset: false,
        label: 'BMC Password',
        apiname: 'secret.hardwaremanagementpassword',
        required: false,
        readonly: false,
        type: 'password',
        placeholder: null
    },
    rack: {
        defvalue: null,
        label: 'Rack',
        apiname: 'location.rack',
        required: false,
        readonly: false,
        type: 'text',
        placeholder: null
    },
    u: {
        defvalue: null,
        label: 'U',
        apiname: 'location.u',
        required: false,
        readonly: false,
        type: 'text',
        placeholder: null
    },
    enclosure: {
        defvalue: null,
        label: 'Enclosure',
        apiname: 'enclosure.manager',
        required: false,
        readonly: false,
        type: 'text',
        placeholder: null
    },
    bay: {
        defvalue: null,
        label: 'Enclosure Bay',
        apiname: 'enclosure.bay',
        required: false,
        readonly: false,
        type: 'text',
        placeholder: null
    },
    model: {
        defvalue: null,
        label: 'Model',
        apiname: 'id.model',
        required: false,
        readonly: false,
        type: 'text',
        placeholder: null
    },
    nodes: {
        defvalue: null,
        label: 'Nodes',
        apiname: 'nodes',
        required: false,
        readonly: false,
        type: 'text',
        placeholder: 'n1, n2, n3'
    }

};

function editGroup() {
    var uidlg = this;
    var inputs = uidlg.getElementsByTagName('input');
    //var jsontext = '{';
    var json = {};
    var groupname = "";
    for (var i=0; i < inputs.length; i++) {
        var field = inputs[i].id;
        var apifield = fields[field].apiname;
        var apival = inputs[i].value;
        var required = inputs[i].required;

        if(apival == null){  // send empty strings, not null values
            apival = "";
        }

        if(apival == "" && required){   // required field is empty, return without saving
            alert("Please enter a value for all required (*) fields.");
            return false;
        }


        if(apifield != null && apifield != "null" && apival != ""){   // create json to send to api and update fields json

            if(field == "nodes"){   // parse through list, convert to array
                var nodesNoSpaces = apival.replace(/ /g, "");   // remove all spaces, so that "n1, n2, n3" and "n1,n2,n3" are treated the same
                var nodesArray = nodesNoSpaces.split(",");  // split into an array
                apival = nodesArray;
            }

            json[apifield] = apival;
        }

        // update fields
        fields[field].defvalue = apival;

    }

    var groupname = document.getElementById("groupname").value;
    var url = "/confluent-api/nodegroups/" + groupname + "/attributes/all";

    putRequest(url, json, refresh);

}


function clearGroupUI(){
    for (var field in fields){
        fields[field].defvalue = null;
    }
}

function editGroupUI() {
    var uidlg = document.createElement('div');      // create dialog div
    uidlg.classList.add('modaldialog');

    var heading = uidlg.appendChild(document.createElement("span"));
    heading.classList.add("noteHeader");
    heading.innerHTML = "Edit Group";

    var dialogContent = document.createElement('div');

    // add content div
    uidlg.appendChild(dialogContent);

    // create table
    var dialogTable = document.createElement('table');
    dialogContent.appendChild(dialogTable);

    // fill content
    //fillDataFromJson(dialogContent, true);
    getNodegroups(dialogTable);

    var bg = document.createElement('div');
    bg.classList.add('modalbg');
    document.body.appendChild(bg);
    bg.appendChild(uidlg);
    uidlg.appendChild(document.createElement('hr'));
    var save = document.createElement('button');
    save.textContent = l('Save');
    save.addEventListener('click', closeDialog.bind([uidlg, bg, editGroup.bind(uidlg)]));
    uidlg.appendChild(save);
    var cancelbtn = document.createElement('button');
    cancelbtn.textContent = l('Cancel');
    cancelbtn.addEventListener('click', closeDialog.bind([uidlg, bg, clearGroupUI]));
    uidlg.appendChild(cancelbtn);
}


function fillDataFromJson(dialog, initial){


    if(initial == null){
        initial = false;
    }

    if(dialog == null){
        dialog = this;
    }


    // clear existing fields
    while(dialog.childNodes.length > 1){
        dialog.removeChild(dialog.childNodes[1]);
    }


    // loop through all fields, defined above. Load info (label and input) into dialog box
    for (var field in fields) {
        if(field != "groupname"){
            var curval = localStorage.getItem(field);   // I don't think this is applicable for new groups since we won't be saving
            if (curval === null) {
                curval = fields[field].defvalue;
            }


            // set up initial requirements
            if(initial == true){

                fields['groupname'].required = true;

            }

            // set up all inputs and labels
            if(fields[field].type == 'select'){
                var curinput = document.createElement('select');
            }else{
               var curinput = document.createElement('input');
               curinput.type = fields[field].type;
               if(fields[field].placeholder != null) curinput.placeholder = fields[field].placeholder;
            }
            
            //curinput.id = fields[field].apiname;
            curinput.id = field;
            if (curval !== null) {
                curinput.value = curval;
            }
            curinput.required = fields[field].required;
            curinput.readOnly = fields[field].readonly;
            var thelab = document.createElement('label');
            thelab.setAttribute('for', curinput.id);
            thelab.textContent = l(fields[field].label) + ':';

            if(curinput.required == true){
                thelab.textContent = "* " + thelab.textContent;
            }

            var newRow = dialog.appendChild(document.createElement('tr'));
            var newLabelTd = newRow.appendChild(document.createElement('td'));
            var newInputTd = newRow.appendChild(document.createElement('td'));

            newLabelTd.appendChild(thelab);
            newInputTd.appendChild(curinput);
            //dialog.appendChild(document.createElement('br'));
        }
    }

}

function getNodegroups(dialog){

    var groupSelect = document.createElement("select");
    groupSelect.id = "groupname";
    groupSelect.addEventListener("change", function(event){getExistingData(dialog, this.value);});

    var blankOpt = groupSelect.appendChild(document.createElement("option"));
    blankOpt.innerHTML = "";
    blankOpt.value = "";
    blankOpt.id = "blankOpt";

    var thelab = document.createElement('label');
    thelab.setAttribute('for', groupSelect.id);
    thelab.textContent = "Group:";

    var newRow = dialog.appendChild(document.createElement("tr"));
    newRow.id = "groupnameSelectRow";
    var newLabelTd = newRow.appendChild(document.createElement("td"));
    var newInputTd = newRow.appendChild(document.createElement("td"));

    newLabelTd.appendChild(thelab);
    newInputTd.appendChild(groupSelect);
    //dialog.appendChild(document.createElement('br'));

    var url = "/confluent-api/nodegroups/";
    getRequest(url, parseNodegroups.bind(dialog));

}

function parseNodegroups(json){

    var dialog = this;

    if ("_requestdone" in json){
        setTimeout(function(){listNodegroups(dialog);}, 100);
    }

    var groupSelect = document.getElementById("groupname");
    for (var group in json){
        
        if(json[group]["href"] != null){
            var name = json[group]["href"].replace(".json", "").replace("/", "");
        }

        var newOpt = document.createElement("option");
        newOpt.id = "groupname_" + name;
        newOpt.value = name;
        newOpt.innerHTML = name;

        groupSelect.appendChild(newOpt);
    }
}

function listNodegroups(dialog){

}

function parseNodegroupData(json){

    var dialog = this;

    if ("_requestdone" in json){
        setTimeout(function(){fillDataFromJson(dialog);}, 100);
    }

   for (var property in json){

        var value = json[property]["value"];
        var isset = json[property]["isset"];

        switch(property){
            case "hardwaremanagement.manager":
                fields["BMC"].defvalue = value;
                break;
            case "secret.hardwaremanagementuser":
                fields["BMCuser"].isset = isset;
                break;
            case "secret.hardwaremanagementpassword":
                fields["BMCpassword"].isset = isset;
                break;
            case "location.rack":
                fields["rack"].defvalue = value;
                break;
            case "location.u":
                fields["u"].defvalue = value;
                break;
            case "enclosure.manager":
                fields["enclosure"].defvalue = value;
                break;
            case "enclosure.bay":
                fields["bay"].defvalue = value;
                break;
            case "id.model":
                fields["model"].defvalue = value;
                break;
            case "nodes":
                var nodes = json["nodes"].join(", ");
                fields["nodes"].defvalue = nodes;
                break;
            default:
                break;
        }  
    }
}

function getExistingData(dialog, groupname){

    if(dialog == null || dialog == "")
        dialog = this[0]
    if(groupname == null || groupname == "")
        groupname = this[1];

    // get rid of blank opt if it exists
    var groupSelect = document.getElementById("groupname");
    if(groupSelect.childNodes[0].id = "blankOpt"){
        groupSelect.removeChild(groupSelect.childNodes[0]);
    }

    var url = "/confluent-api/nodegroups/" + groupname + "/attributes/all";

    getRequest(url, parseNodegroupData.bind(dialog));

}


var refreshTimeout = 0;

function refresh(){

    clearGroupUI();

    var hash = window.location.hash;

    if(hash == "#compactoverview" || hash == ""){

        if(refreshTimeout == 0){
            refreshTimeout = setTimeout(function(){window.location.reload(true)}, 1);
        }
    }
    
}

window.editGroupUI = editGroupUI;
})();

