// Copyright 2016 Lenovo
;(function() {
'use strict';

// Define all fields
var fields = {
    nodename: {
        defvalue: null,
        label: 'Node Name',
        apiname: 'name',
        required: true,
        readonly: true,
        type: 'text'
    },
    BMC: {
        defvalue: null,
        label: 'BMC',
        apiname: 'hardwaremanagement.manager',
        required: true,
        readonly: false,
        type: 'text'
    },
    BMCuser: {
        defvalue: null,
        isset: false,
        label: 'BMC Username',
        apiname: 'secret.hardwaremanagementuser',
        isset: false,
        required: true,
        readonly: false,
        type: 'text'
    },
    BMCpassword: {
        defvalue: null,
        isset: false,
        label: 'BMC Password',
        apiname: 'secret.hardwaremanagementpassword',
        isset: false,
        required: true,
        readonly: false,
        type: 'password'
    },
    rack: {
        defvalue: null,
        label: 'Rack',
        apiname: 'location.rack',
        required: false,
        readonly: false,
        type: 'text'
    },
    u: {
        defvalue: null,
        label: 'U',
        apiname: 'location.u',
        required: false,
        readonly: false,
        type: 'text'
    },
    enclosure: {
        defvalue: null,
        label: 'Enclosure',
        apiname: 'enclosure.manager',
        required: false,
        readonly: false,
        type: 'text'
    },
    bay: {
        defvalue: null,
        label: 'Enclosure Bay',
        apiname: 'enclosure.bay',
        required: false,
        readonly: false,
        type: 'text'
    },
    model: {
        defvalue: null,
        label: 'Model',
        apiname: 'id.model',
        required: false,
        readonly: false,
        type: 'text'
    },
    serial: {
        defvalue: null,
        label: 'Serial',
        apiname: 'id.serial',
        required: false,
        readonly: false,
        type: 'text'
    },
    MAC: {
        defvalue: null,
        label: 'MAC',
        apiname: null,
        required: false,
        readonly: false,
        type: 'text'
    },
    groups: {
        defvalue: null,
        label: 'Nodegroups',
        apiname: 'groups',
        required: false,
        readonly: false,
        type: 'text',
        placeholder: 'group1, group2, group3'
    }

};

function editNode() {
    var uidlg = this;
    var inputs = uidlg.getElementsByTagName('input');
    //var jsontext = '{';
    var json = {};
    for (var i=0; i < inputs.length; i++) {
        var field = inputs[i].id;
        var apifield = fields[field].apiname;
        var apival = inputs[i].value;
        var required = inputs[i].required;

        if(apival == null){  // send empty strings, not null values
            apival = "";
        }

        if(apival == "" && required){   // required field is empty, return without saving
            alert("Please enter a value for all required (*) fields.");
            return false;
        }


        if(apifield != null && apifield != "null" && apifield != "name" && apival != ""){   // create json to send to api and update fields json
            
            if(apifield == "groups"){
                var groups = apival.replace(", ", ",").split(",");
                apival = groups;
            }

            json[apifield] = apival;
        }

        // update fields
        fields[field].defvalue = apival;

    }

    var nodename = nodeInfo["nodename"];

    putRequest('/confluent-api/nodes/' + nodename + "/attributes/all", json, refresh);

}



var nodeInfo = new Object();

function compileNode(data){

    for (var item in data){

        if(item == '_requestdone'){
            return;
        }

        var property = item;
        var value = data[item];

        nodeInfo[property] = value;

    }

}




// 


function clearNodeUI(){
    for (var field in fields){
        fields[field].defvalue = null;
    }
}

function editNodeUI(nodename) {

    if(nodename == null || nodename == ""){
        nodename = this;
    }

    fields["nodename"].defvalue = nodename;

    var uidlg = document.createElement('div');      // create dialog div
    uidlg.classList.add('modaldialog');

    var heading = uidlg.appendChild(document.createElement("span"));
    heading.classList.add("noteHeader");
    heading.innerHTML = "Edit Node";
    // var headBreak = uidlg.appendChild(document.createElement("span"));
    // headBreak.innerHTML = "<br/>"; 

    var dialogContent = document.createElement('div');

    // add content div
    uidlg.appendChild(dialogContent);

    // fill content
    //fillDataFromJson(dialogContent, true);
    getNodeData(nodename, dialogContent);

   

    var bg = document.createElement('div');
    bg.classList.add('modalbg');
    document.body.appendChild(bg);
    bg.appendChild(uidlg);
    uidlg.appendChild(document.createElement('hr'));
    var save = document.createElement('button');
    save.textContent = l('Save');
    save.addEventListener('click', closeDialog.bind([uidlg, bg, editNode.bind(uidlg)]));
    uidlg.appendChild(save);
    var cancelbtn = document.createElement('button');
    cancelbtn.textContent = l('Cancel');
    cancelbtn.addEventListener('click', closeDialog.bind([uidlg, bg, clearNodeUI]));
    uidlg.appendChild(cancelbtn);
}


function fillDataFromJson(dialog, initial){


    if(initial == null){
        initial = false;
    }

    if(dialog == null){
        dialog = this;
    }

    dialog.innerHTML = "";

    // loop through all fields, defined above. Load info (label and input) into dialog box
    for (var field in fields) {
        var curval = localStorage.getItem(field);   // I don't think this is applicable for new nodes since we won't be saving
        if (curval === null) {
            curval = fields[field].defvalue;
        }


        // set up initial requirements
        if(initial == true){

            fields['nodename'].required = true;
            fields['BMC'].required = true;
            fields['BMCuser'].required = true;
            fields['BMCpassword'].required = true;


        }

        // set up all inputs and labels
        var curinput = document.createElement('input');
        //curinput.id = fields[field].apiname;
        curinput.id = field;
        if (curval !== null) {
            curinput.value = curval;
            nodeInfo[field] = curval;
        }
        curinput.required = fields[field].required;
        curinput.readOnly = fields[field].readonly;
        curinput.type = fields[field].type;
        if(fields[field]["placeholder"] != null) curinput.placeholder = fields[field]["placeholder"];
        var thelab = document.createElement('label');
        thelab.setAttribute('for', curinput.id);
        thelab.textContent = l(fields[field].label) + ':';

        if(curinput.required == true){
            thelab.textContent = "* " + thelab.textContent;
        }

        dialog.appendChild(thelab);
        dialog.appendChild(curinput);
        dialog.appendChild(document.createElement('br'));
    }

}



function getNodeData(node, dialog){

    getRequest("/confluent-api/nodes/" + node + "/attributes/all", parseNodeData.bind(dialog));

}

function parseNodeData(json){

    var dialog = this;

    if ("_requestdone" in json){
        setTimeout(function(){fillDataFromJson(dialog);}, 100);
    }

    for (var property in json){

        var value = json[property]["value"];
        var isset = json[property]["isset"];

        switch(property){
            case "name":
                fields["nodename"].defvalue = value;
                break;
            case "hardwaremanagement.manager":
                fields["BMC"].defvalue = value;
                break;
            case "secret.hardwaremanagementuser":
                fields["BMCuser"].isset = isset;
                console.log(isset);
                if(isset){
                    fields["BMCuser"].required = false;
                }
                break;
            case "secret.hardwaremanagementpassword":
                fields["BMCpassword"].isset = isset;
                console.log(isset);
                if(isset){
                    fields["BMCpassword"].required = false;
                }
                break;
            case "location.rack":
                fields["rack"].defvalue = value;
                break;
            case "location.u":
                fields["u"].defvalue = value;
                break;
            case "enclosure.manager":
                fields["enclosure"].defvalue = value;
                break;
            case "enclosure.bay":
                fields["bay"].defvalue = value;
                break;
            case "model":
                fields["model"].defvalue = value;
                break;
            case "serial":
                fields["serial"].defvalue = value;
                break;
            case "MAC":
                fields["MAC"].defvalue = value;
                break;
            case "groups":
                fields["groups"].defvalue = json[property].join(", ");
                break;
            default:
                break;
        }

  
    }

}




var refreshTimeout = 0;

function refresh(){

    clearNodeUI();

    var hash = window.location.hash;

    if(hash == "#compactoverview" || hash == ""){

        if(refreshTimeout == 0){
            refreshTimeout = setTimeout(function(){window.location.reload(true)}, 1);
        }
    }
    
}

window.editNodeUI = editNodeUI;
})();

