;(function() {
'use strict';
var mainpane;
var active = true;
var headingpane;
var hardwarepane;
var hardwarelogpane;
var firmwarepane;
var temperaturepane;
var hardwarelogbuttonpane;
var firmwarebuttonpane;
var temperaturebuttonpane;
var hardwareLogEvents = [];
var eventIndex = 0;
var powerState;
var pendingPowerStateUpdate = null;
var pendingHardwareLogUpdate = null;
var pendingHealthUpdate = null;
var clearLogButton;
var refreshLogButton;
var filterDropDown;
var filterLabel;
var changePagesSelect;
var firstPageButton;
var lastPageButton;
var nextPageButton;
var prevPageButton;

	/* General Use Functions */

	function getNodeName(){
	  var url = window.location.hash;
	  var nodeName = url.replace("#nodeview:", "");
	  return nodeName;
	}

	function fetchData(table){
		switch(table){
			case "hardwareInfo":
				getHardwareInfo(getNodeName());
				break;
			case "firmwareInfo":
				getFirmwareInfo(getNodeName());
				break;
			case "hardwareLog":
				getHardwareLog(getNodeName());
				break;
			default:
				break;
		}
	}


    function hide(div, hide){
    	if(hide){
    		div.classList.add("hidden");
    	}else{
    		div.classList.remove("hidden");
    	}
    }

    function getHealthIcon(healthStr){

	    var healthicon = document.createElement('i');

	    switch(healthStr){
	        case "warning":
	            healthicon.setAttribute('class', 'fa fa-exclamation-triangle yellow');
	            break;
	        case "failed":
	        case "critical":
	            healthicon.setAttribute('class', 'fa fa-times-circle red');
	            break;
	        case "ok":
	            healthicon.setAttribute('class', 'fa fa-check-square-o green');
	            break;
	        case "unknown":
	        default:
	            healthicon.setAttribute('class', 'fa fa-question');
	            break;
	    }

	    return healthicon;

	}

    /* End of General Use Functions */

	/* Table Functions */

	function makeTable(tableID, pane, columns, widths){

		// Create Table
		var table = pane.appendChild(document.createElement("table"));
		table.id = tableID;
		table.border = "1";
		table.setAttribute("class", "nodeViewTable");

		// Create Head
		table.head = table.appendChild(document.createElement("thead"));
		var headRow = table.head.appendChild(document.createElement("tr"));

		// Create Foot
		table.foot = table.appendChild(document.createElement("tfoot"));
		var footRow = table.foot.appendChild(document.createElement("tr"));

		// Make each column's head and foot
		var i = 0;
		var th, td;
		for(i = 0; i < columns.length; i++){
			// Head
			th = headRow.appendChild(document.createElement("th"));
			th.setAttribute("class", "nodeViewTh");
			th.style.width = widths[i];
			th.innerHTML = columns[i];

			// Foot
			td = footRow.appendChild(document.createElement("td"));
			var spinspan = td.appendChild(document.createElement('span'));
			spinspan.classList.add('icon', 'fa', 'fa-spinner', 'fa-spin');
			td.setAttribute("class", "centered");

			// hide first column of hardware log
			if(tableID == "hardwareLogTable" && i == 0){ 
				th.setAttribute("class", "hidden");
				td.setAttribute("class", "hidden");
			}
		}

		// Create Body
		table.tableBody = table.appendChild(document.createElement("tbody"));

	}

	function addTableRow(tableID, data, classes){
	
		// Get table and tablebody
		var table = document.getElementById(tableID);
		var body = table.tableBody;

		// Create new row
		var newRow = body.appendChild(document.createElement("tr"));
		var i = 0;
		var td;
		for(i = 0; i < data.length; i++){
			td = newRow.appendChild(document.createElement("td"));
			td.setAttribute("class", classes[i]);
			td.innerHTML = data[i];
		}
	}

	/* End Table Functions */

	/* Button Functions */

	function addButton(buttonID, pane, associatedPane, text, forTable){
		var button = pane.appendChild(document.createElement('button'));
		button.id = buttonID;
		button.setAttribute("class", "nodeViewButton");
		button.innerHTML = "<span class='fa fa-caret-right'></span> &nbsp;"+l(text);

		button.addEventListener("click", function(){
			if(associatedPane.className == "hidden"){
				hide(associatedPane, false);
				button.innerHTML = "<span class='fa fa-caret-down'></span> "+l(text);
				if(!associatedPane.fetched){
					fetchData(forTable);
	        	}

	        	if(associatedPane == hardwarelogpane){
	        		hide(clearLogButton, false);
	        		hide(refreshLogButton, false);
	        		hide(filterLabel, false);
	        		hide(filterDropDown, false);
	        	}
			}else{
	        	hide(associatedPane, true);
	        	button.innerHTML = "<span class='fa fa-caret-right'></span> &nbsp;"+l(text);

	        	if(associatedPane == hardwarelogpane){
	        		hide(clearLogButton, true);
	        		hide(refreshLogButton, true);
	        		hide(filterLabel, true);
	        		hide(filterDropDown, true);
	        	}
	        }
		});
	}

	/* End Button Functions */


	/* Hardware Log Functions */

	function parseHardwareLog(json){
		var hardwareLogTable = document.getElementById('hardwareLogTable');

        if (json._requestdone) {
            hardwareLogTable.removeChild(hardwareLogTable.foot);
            paginate(hardwareLogTable, 10);            
        }

		if(json.events != null){
			var events = json.events;
			events.reverse(); // reverse order = most recent first
			events.forEach(function(eventItem){
				var severity = eventItem.severity;
				var healthIcon = getHealthIcon(severity).outerHTML;
				var timestamp = eventItem.timestamp.replace("T", " ");
				var componentType = eventItem.component_type;
				var component = eventItem.component;
				var event = eventItem.event;

				addTableRow("hardwareLogTable", [severity, healthIcon, timestamp, componentType, component, event], ["hidden", "nodeViewTd centered", "nodeViewTd", "nodeViewTd", "nodeViewTd", "nodeViewTd"]);

			});
		}
	}

	function getHardwareLog(nodeName){
		getRequest("/confluent-api/nodes/" + nodeName + "/events/hardware/log", parseHardwareLog);
		hardwarelogpane.fetched = true;
	}

	function paginate(table, perPage){

		hardwareLogEvents = [];
		eventIndex = 0;
		[].forEach.call(table.tableBody.children, function(row){
			hardwareLogEvents.push(row);
		});

		// Add page buttons
		var pageButtons = hardwarelogpane.appendChild(document.createElement("div"));
		pageButtons.setAttribute("class", "centered");

		// First page button
		firstPageButton = pageButtons.appendChild(document.createElement("button"));
		firstPageButton.setAttribute("class", "pageButton centered");
		firstPageButton.innerHTML = "<span class='fa fa-caret-left'></span><span class='fa fa-caret-left'></span> " + l("First Page");
		firstPageButton.addEventListener("click", function(){
			eventIndex = 0;
			prevPageButton.disabled = true;
			firstPageButton.disabled = true;
			if(perPage <= hardwareLogEvents.length){
				nextPageButton.disabled = false;
				lastPageButton.disabled = false;
			}
			showEvents(eventIndex, perPage);
		});

		// Previous page button
		prevPageButton = pageButtons.appendChild(document.createElement("button"));
		prevPageButton.setAttribute("class", "pageButton centered");
		prevPageButton.innerHTML = "<span class='fa fa-caret-left'></span> " + l("Previous Page");
		prevPageButton.addEventListener("click", function(){
			eventIndex = (eventIndex - perPage) >= 0 ? (eventIndex - perPage) : 0;
			if(eventIndex == 0){
				prevPageButton.disabled = true;
				firstPageButton.disabled = true;
			}
			if(perPage <= hardwareLogEvents.length){
				nextPageButton.disabled = false;
				lastPageButton.disabled = false;
			}
			showEvents(eventIndex, perPage);
		});

		//Change pages per side
		var changePagesText = pageButtons.appendChild(document.createElement("label"));
		changePagesText.setAttribute("class", "nodeViewLabel");
		changePagesText.innerHTML = "Events Per Page: ";

		changePagesSelect = pageButtons.appendChild(document.createElement("select"));
		var option5 = changePagesSelect.appendChild(document.createElement("option"));
		var option10 = changePagesSelect.appendChild(document.createElement("option"));
		var option15 = changePagesSelect.appendChild(document.createElement("option"));
		var option20 = changePagesSelect.appendChild(document.createElement("option"));

		option5.innerHTML = "5";
		option5.value = "5";

		option10.innerHTML = "10";
		option10.value = "10";
		option10.selected = true;

		option15.innerHTML = "15";
		option15.value = "15";

		option20.innerHTML = "20";
		option20.value = "20";

		changePagesSelect.addEventListener("change", function(){
			perPage = parseInt(this.value, 10);
			showEvents(eventIndex, perPage);
			if(perPage <= hardwareLogEvents.length){
				nextPageButton.disabled = false;
				lastPageButton.disabled = false;
			}
		});



		// Next page button
		nextPageButton = pageButtons.appendChild(document.createElement("button"));
		nextPageButton.setAttribute("class", "pageButton centered");
		nextPageButton.innerHTML = l("Next Page") + " <span class='fa fa-caret-right'></span>";
		nextPageButton.addEventListener("click", function(){
			eventIndex = (eventIndex + perPage) < hardwareLogEvents.length ? (eventIndex + perPage) : eventIndex;
			if(eventIndex + perPage >= hardwareLogEvents.length){
				nextPageButton.disabled = true;
				lastPageButton.disabled = true;
			}
			firstPageButton.disabled = false;
			prevPageButton.disabled = false;
			showEvents(eventIndex, perPage);
		});

		// Last page button
		lastPageButton = pageButtons.appendChild(document.createElement("button"));
		lastPageButton.setAttribute("class", "pageButton centered");
		lastPageButton.innerHTML = l("Last Page") + " <span class='fa fa-caret-right'></span><span class='fa fa-caret-right'></span>";
		lastPageButton.addEventListener("click", function(){
			var lastPageNumber = Math.floor(hardwareLogEvents.length / perPage);
			eventIndex = lastPageNumber * perPage;

			lastPageButton.disabled = true;
			nextPageButton.disabled = true;
			firstPageButton.disabled = false;
			prevPageButton.disabled = false;

			showEvents(eventIndex, perPage);
		});

		firstPageButton.disabled = true;
		prevPageButton.disabled = true;
		if(perPage > hardwareLogEvents.length){
			nextPageButton.disabled = true;
			lastPageButton.disabled = true;
		}else{
			nextPageButton.disabled = false;
			lastPageButton.disabled = false;
		}

		showEvents(0, perPage);
	}

	function showEvents(start, perPage){
		hardwareLogEvents.forEach(function(row, index){
			if(index >= start && index < (start + perPage)){
				row.style.display = "table-row";
			}else{
				row.style.display = "none";
			}
		});
	}

	function filterHardwareLog(table, filter){
		// deal with null filter by including all
		if(filter == null){
			filter = ['warning', 'failed', 'critical', 'ok', 'unknown'];
		}

		// get appropriate list of events
		hardwareLogEvents = [];
		[].forEach.call(table.tableBody.children, function(row){
			var health = row.children[0].innerHTML;
			if(filter.indexOf(health) >= 0){   //if this health is included in filter
				hardwareLogEvents.push(row);
			}
			row.style.display = "none"; // hide all rows
		});

		var perPage = changePagesSelect.value;
		showEvents(0, perPage);

		firstPageButton.disabled = true;
		prevPageButton.disabled = true;
		if(perPage > hardwareLogEvents.length){
			nextPageButton.disabled = true;
			lastPageButton.disabled = true;
		}else{
			nextPageButton.disabled = false;
			lastPageButton.disabled = false;
		}
	}

	function clearHardwareLog(){
		deleteRequest("/confluent-api/nodes/" + getNodeName() + "/events/hardware/log", updateHardwareLog);
	}

	function updateHardwareLog(json){
		if(json){
			if(json._requestdone){
				removeTable(hardwarelogpane);
				makeTable("hardwareLogTable", hardwarelogpane, ["Severity Hidden", "Severity", "Timestamp", "Component Type", "Component", "Event"], ["0%", "10%", "15%", "25%", "20%", "30%"]);
				pendingHardwareLogUpdate = setTimeout(function(){   // wait for clear log to show up in log
					fetchData("hardwareLog");
					updateHealthInfo();
				}, 5000);
			}
		}else{ // generic updateHardwareLog
			removeTable(hardwarelogpane);
			makeTable("hardwareLogTable", hardwarelogpane, ["Severity Hidden", "Severity", "Timestamp", "Component Type", "Component", "Event"], ["0%", "10%", "15%", "25%", "20%", "30%"]);
			fetchData("hardwareLog");
		}
	}

	function removeTable(pane){
		pane.innerHTML = "";
	}

	function createFilter(filterChoice){
		var filter = [];
		switch (filterChoice){
			case "critical":
				filter = ['critical', 'failed', 'unknown'];
				break;
			case "warning":
				filter = ['critical', 'failed', 'warning', 'unknown'];
				break;
			case "all":
			default:
				filter = ['critical', 'failed', 'warning', 'unknown', 'ok'];
				break;
		}
		return filter;
	}
	/* End of Hardware Log Functions */

	/* Firmware Table Functions */

	function parseFirmwareInfo(json){
		var firmwareTable = document.getElementById('firmwareTable');
        if (json._requestdone) {
            firmwareTable.removeChild(firmwareTable.foot);            
        }

		if(json.firmware != null){
			json.firmware.forEach(function(firmwareItem){
				var firmwareName = Object.keys(firmwareItem)[0];
				var build = firmwareItem[firmwareName].build == null ? "" : firmwareItem[firmwareName].build;
				var date = firmwareItem[firmwareName].date == null ? "" : firmwareItem[firmwareName].date.substring(0,10);
				var version = firmwareItem[firmwareName].version == null ? "" : firmwareItem[firmwareName].version;

				addTableRow("firmwareTable", [firmwareName, build, date, version], ["nodeViewTd firmware", "nodeViewTd centered", "nodeViewTd centered", "nodeViewTd centered"]);
			});
		}
	}

	function getFirmwareInfo(nodeName){
		getRequest("/confluent-api/nodes/" + nodeName + "/inventory/firmware/all/all", parseFirmwareInfo);
		firmwarepane.fetched = true;
	}

	/* End of Firmware Table Functions */

	/* Hardware Info Functions */

	function parseHardwareInfo(json){
		//var hardwareTable = document.getElementById('hardwareTable');
        if (json._requestdone) {
        //    hardwareTable.removeChild(hardwareTable.foot);            
        	var healthDiv = document.getElementById("healthDiv");
        	hide(healthDiv, false);
        }

		if(json.inventory != null){
			json.inventory.forEach(function(hardwareItem){
				var productName = hardwareItem.information["Product name"];
				var serialNumber = hardwareItem.information["Serial Number"];

				//addTableRow("hardwareTable", [productName, serialNumber], ["nodeViewTd centered", "nodeViewTd centered"]);
				var heading = document.getElementById("heading");
				heading.innerHTML += ": " + productName + " (SN: " + serialNumber + ")";
			});
		}
	}

	function getHardwareInfo(nodeName){
		getRequest("/confluent-api/nodes/" + nodeName + "/inventory/hardware/all/system", parseHardwareInfo);
	}

	/* End of Hardware Info Functions */

	/* Health Functions */

	function getHealthInfo(nodeName){
		if(!nodeName){
			nodeName = getNodeName();
		}
		getRequest("/confluent-api/nodes/" + nodeName + "/health/hardware", parseHealthInfo);
		pendingHealthUpdate = setTimeout(getHealthInfo, 30000);
	}

	function parseHealthInfo(json){
		var healthDiv = document.getElementById("healthDiv");
		var healthVal = null;

		if(json.health != null){
			healthVal = json.health.value;
			healthDiv.innerHTML = "Health: ";
			healthDiv.appendChild(getHealthIcon(healthVal));
		}

		if (json.sensors != null){
			[].forEach.call(json.sensors, function(sensor, index){
				if(index > 0){
					healthDiv.innerHTML += "<br/>";
				}
				healthDiv.innerHTML += (" " + sensor.name + ": " + sensor.states.join(", "));
			});
		}
	}

	function updateHealthInfo(){
		var healthDiv = document.getElementById("healthDiv");
		healthDiv.innerHTML = "Health: ";
		getHealthInfo(getNodeName());
	}
	/* End of Health Functions */


	/* Setup functions */

	function createPanes(){

		// Create panes
		headingpane = mainpane.appendChild(document.createElement('div'));
		hardwarepane = mainpane.appendChild(document.createElement('div'));
        firmwarebuttonpane = mainpane.appendChild(document.createElement('div'));
        firmwarepane = mainpane.appendChild(document.createElement('div'));
        hardwarelogbuttonpane = mainpane.appendChild(document.createElement('div'));
        hardwarelogpane = mainpane.appendChild(document.createElement('div'));
        temperaturebuttonpane = mainpane.appendChild(document.createElement('div'));
        temperaturepane = mainpane.appendChild(document.createElement('div'));

        // Initialize fetched attributes to false
        firmwarepane.fetched = false;
        hardwarelogpane.fetched = false;
        temperaturepane.fetched = false;

        // Hide all panes that default as hidden
        hide(firmwarepane, true);
       	hide(hardwarelogpane, true);
       	hide(temperaturepane, true);

       	// Set classes for panes
       	headingpane.setAttribute("class", "centered");
       	firmwarebuttonpane.setAttribute("class", "buttonPane");
        hardwarelogbuttonpane.setAttribute("class", "buttonPane");
        temperaturebuttonpane.setAttribute("class", "buttonPane");
	}

	function createButtons(){
		addButton("firmwareButton", firmwarebuttonpane, firmwarepane, "Firmware", "firmwareInfo");
        addButton("hardwareLogButton", hardwarelogbuttonpane, hardwarelogpane, "Hardware Log", "hardwareLog");
        addButton("temperatureButton", temperaturebuttonpane, temperaturepane, "Temperature and Power", "temperature");

        clearLogButton = hardwarelogbuttonpane.appendChild(document.createElement('button'));
        clearLogButton.innerHTML = "<span class='fa fa-remove'></span> &nbsp;" + l("Clear Log");
        clearLogButton.setAttribute("class", "hardwareLogButton");
        clearLogButton.addEventListener("click", function(){
        	confirmAction("Are you sure you want to clear the log?", clearHardwareLog);
        });
        hide(clearLogButton, true);

        refreshLogButton = hardwarelogbuttonpane.appendChild(document.createElement('button'));
        refreshLogButton.innerHTML = "<span class='fa fa-refresh'></span> &nbsp;" + l("Refresh Log");
        refreshLogButton.setAttribute("class", "hardwareLogButton");
        refreshLogButton.addEventListener("click", function(){
        	updateHardwareLog();
        });
        hide(refreshLogButton, true);

        filterLabel = hardwarelogbuttonpane.appendChild(document.createElement('label'));
        filterLabel.innerHTML = "&nbsp;&nbsp;&nbsp;Filter: ";
        filterLabel.setAttribute("class", "hardwareLogButton");
        filterDropDown = hardwarelogbuttonpane.appendChild(document.createElement('select'));
        var criticalOption = filterDropDown.appendChild(document.createElement('option'));
        var warningOption = filterDropDown.appendChild(document.createElement('option'));
        var allOption = filterDropDown.appendChild(document.createElement('option'));
        criticalOption.innerHTML = "Critical";
        criticalOption.value = "critical";
        warningOption.innerHTML = "Warning";
        warningOption.value = "warning";
        allOption.innerHTML = "All";
        allOption.value = "all";
        allOption.selected = true;
        filterDropDown.addEventListener("change", function(){
        	filterHardwareLog(hardwareLogTable, createFilter(filterDropDown.value));
        });
        hide(filterLabel, true);
        hide(filterDropDown, true);
	}

	function createTables(){
		//makeTable("hardwareTable", hardwarepane, ["Name", "Serial Number"], ["50%", "50%"]);
        makeTable("hardwareLogTable", hardwarelogpane, ["Severity Hidden", "Severity", "Timestamp", "Component Type", "Component", "Event"], ["0%", "10%", "15%", "25%", "20%", "30%"]);    
        makeTable("firmwareTable", firmwarepane, ["Firmware", "Build", "Date", "Version"], ["40%", "20%", "20%", "20%"]);
        //makeTable("temperatureTable", temperaturepane, ["Health", "Name", "Temperature"], ["20%", "40%", "40%"]);
	}

	function createHeading(){
		var heading = headingpane.appendChild(document.createElement("div"));
		heading.id = "heading";
		heading.setAttribute("class", "centered nodeViewHeading");
		heading.innerHTML = getNodeName();
		//fetchData("hardwareInfo");
		getHardwareInfo(getNodeName());
	}

	function createHealth(){
		var health = headingpane.appendChild(document.createElement("div"));
		health.setAttribute("class", "centered");
		//health.innerHTML = "Health: ";
		health.id = "healthDiv";
		hide(health, true);
		getHealthInfo(getNodeName());
	}


	/* End of Setup Functions */

	/* Toolbar Functions */

	function maketoolbar() {
	    var toolrow = document.createElement('ul');
	    toolrow.classList.add('toolbar', 'header');

	    // get current power state
	    updatePowerState();

	    var powerMenu = addToolMenu2(toolrow, 'Power', 'power-off', null, 'powerStateIcon'); 
	    var powerOffMenu = addToolMenu(powerMenu, 'Off', 'power-off', 'dropright');
	    addMenuItem(powerOffMenu, 'Shutdown', globalPower.bind('shutdown'), 'bed');
	    addMenuItem(powerOffMenu, 'Force Off', globalPower.bind('off'), 'power-off');
	    var powerOnMenu = addToolMenu(powerMenu, 'Boot', 'power-off', 'dropright');
	    addMenuItem(powerOnMenu, 'Boot', globalPower.bind('boot'), 'refresh');
        addMenuItem(powerOnMenu, 'Boot to Setup', bootOptions.bind('setup'), 'refresh');
        addMenuItem(powerOnMenu, 'Boot to Network', bootOptions.bind('network'), 'refresh');
        addMenuItem(powerOnMenu, 'Boot to Media', bootOptions.bind('cd'), 'refresh');

	    var identifyMenu = addToolMenu(toolrow, 'Identify', 'lightbulb-o');
	    addMenuItem(identifyMenu, 'Activate', identify.bind('on'), 'check');
	    addMenuItem(identifyMenu, 'Deactivate', identify.bind('off'), 'times');

	    var terminalsMenu = addToolMenu(toolrow, 'Terminals', 'desktop');
	    addMenuItem(terminalsMenu, 'Open Console', openTerminal.bind('console'), 'desktop');
	    addMenuItem(terminalsMenu, 'Open Shell', openTerminal.bind('shell'), 'terminal');
			var guiitem = addMenuItem(toolrow, 'Open Management Controller Web GUI', openGui, 'share');
			guiitem.classList.add('navitem');

	    return toolrow;
	}

	function globalPower(e) {
        if (e && !e.bootmode && !e.target) { return; }
	    var targets = [getNodeName()];
	    changePower(targets, this);
	}

	function identify(){
		var data = { 'identify': this };
        var url = "/confluent-api/nodes/" + getNodeName() + "/identify";
        postRequest(url, data, donull);
	}

	function openTerminal(){
		var type = this;
		new TerminalWindow("/confluent-api/nodes/" + getNodeName() + "/" + type + "/sessions/", getNodeName() + ' ' + l(type), undefined, undefined, quickIcon(getNodeName(), true, true));
	}
	function openGui() {
		window.open("/confluent-api/nodes/" + getNodeName() + "/forward/web");
	}
	function bootOptions(){
        var nextdevice = this;
        var data = { "bootmode": "uefi", "nextdevice": nextdevice, "persistent": "False" };
        var url = "/confluent-api/nodes/" + getNodeName() + "/boot/nextdevice";
        postRequest(url, data, globalPower.bind('boot'));
    };

    function updatePowerState(){
    	var url = "/confluent-api/nodes/" + getNodeName() + "/power/state";
	    getRequest(url, updatePowerIcons);
	    pendingPowerStateUpdate = setTimeout(updatePowerState, 30000);
    }

    function updatePowerIcons(data){
    	if (data.state != null){
    		powerState = data.state.value;
	    	[].forEach.call(document.getElementsByClassName('powerStateIcon'), function(i){
	    		if(powerState == "on")
	    			i.classList.add('poweron');
	    		else
	    			i.classList.remove('poweron');
	    	});
    	}

    	// if("_requestdone" in data){
    	// 	pendingPowerStateUpdate = setTimeout(updatePowerState, 2000);
    	// }
    }

    /* End of Toolbar Functions */

    
    /* Nodeview Functions */

    function clearPendingUpdates(){
    	pendingPowerStateUpdate = null;
    	pendingHealthUpdate = null;
    	pendingHardwareLogUpdate = null;
    }

    function resizeCharts() {
        [].forEach.call(document.getElementsByClassName('js-plotly-plot'),
            function(e) {
                e.style.height = (0.8 * window.innerHeight) + 'px';
                e.style.width = (0.49 * window.innerWidth) + 'px';
                Plotly.Plots.resize(e);
            });
    }

    function closenodeview() {
    	window.removeEventListener('resize', resizeCharts);
    	clearPendingUpdates();
    }

    function nodeview() {
    	var nodename;
    	if((nodename = this) == null){
    		nodename = getNodeName();
    	}
        clear_main();
        clearPendingUpdates(); 
        active = true;
        window.main = nodeview;
        window.closemain = closenodeview;
        var truemainpane = document.getElementById('main');
        mainpane = document.createElement('div');
        mainpane.appendChild(maketoolbar());
        truemainpane.appendChild(mainpane);     

        window.addEventListener('resize', resizeCharts);

        createPanes();
        createHeading();     
       	createButtons();
       	createTables();
       	createHealth(); 	
        
        initpowerchart(temperaturepane);
    };

    /* End of Nodeview Functions */

    window.nodeview = nodeview;
    window.getNodeName = getNodeName;
})();
