;(function() {
'use strict';
var curroverlay = null;
var minvalue;
var maxvalue;
var racku = 42;
var weightedambient = [
    [17, 240],
    [22.5, 120],
    [31.5, 60],
    [45, 0],
];
var overlaysensors = {
    'Ambient Temperature': {
        'sensors': [ 'ambient_temp' ],
        'scale': weightedambient,
    },
    'CPU Temperature': {
        'sensors': [ 'cpu1_temp', 'cpu2_temp' ],
        'scale': null,
    },
    'Power': {
        'sensors': [ 'avg_power' ],
        'scale': null,
    },
};
var overlayselect;

function createRack(rackname) {
    var nrack = document.createElement('div');
    if (rackname != ' ') {
        var namelabel = document.createElement('span');
        namelabel.style.position = 'absolute';
        namelabel.style.top = '100%';
        namelabel.style.marginTop = '0.5em';
        namelabel.appendChild(document.createTextNode('Rack {}'.l(rackname)));
        nrack.appendChild(namelabel);
    }
    nrack.classList.add('rack');
    nrack.style.position = 'relative';
    nrack.style.marginBottom = '1.2pc';
    mainpane.appendChild(nrack);
    return nrack;

}
var rows;
var enclosures;
var nodelocations = {};
var nodedivs;

function setoverlay() {
    if (curroverlay === overlaysensors[this]) { return; }
    curroverlay = overlaysensors[this];
    overlayselect.header.labeltext.textContent = 'Overlay: {} '.l(this);
    if (fetchdata) {
            clearTimeout(fetchdata);
            fetchdata = null;
    }
    for (var nname in nodedivs) {
        nodedivs[nname].getElementsByTagName('span')[1].textContent = "";
        nodedivs[nname].style.background = '#ffffff';
    }
    minvalue = null;
    maxvalue = null;
    colorNodes();
}

function maketoolbar() {
    var toolrow = document.createElement('ul');
    toolrow.classList.add('toolbar', 'header');
    toolrow.appendChild(maketoollabel(l('Noderange:')));
    var nrinput = document.createElement('input');
    nrinput.classList.add('tool');
    nrinput.addEventListener('input', newnoderange.bind(nrinput));
    nrinput.value = noderange;
    toolrow.appendChild(nrinput);
    //var pausebutton = document.createElement('button');
    //pausebutton.textContent = l('Pause');
    //pausebutton.addEventListener('click', pause.bind(pausebutton));
    //toolrow.appendChild(pausebutton);
    nrinput.value = noderange;
    overlayselect = addToolMenu(toolrow, l('Overlay'), 'tachometer');
    for (var lab in overlaysensors) {
        addMenuItem(overlayselect, l(lab), setoverlay.bind(lab));
    }
    return toolrow;
}
var nrchanger;
function newnoderange() {
    var sel
    if (nrchanger) {
        clearTimeout(nrchanger);
    }
    var nnr = this.value;
    
    nrchanger = setTimeout(function() {
        noderange = nnr;
        while (mainpane.firstChild) {
            mainpane.removeChild(mainpane.firstChild);
        }
        if (isinvalidnoderange(nnr)) {
            var errlabel = document.createElement('span');
            errlabel.style.color = 'red';
            errlabel.innerText = 'Invalid noderange syntax';
            mainpane.appendChild(errlabel);
            return;
        }
        getNodeLocations();
    }, 1000);
}
function generateColor(scale, cur) {
    var idx;
    if (scale === null) {
        scale = [[minvalue, 240], [maxvalue, 0]];
    }
    for (idx=0; idx < (scale.length - 1); idx++) {
        if (scale[idx][0] == null || cur < scale[idx+1][0]) {
            break;
        }
    }
    var maxent;
    if (idx == scale.length - 1) {
        maxent = scale[idx];
    } else {
        maxent = scale[idx+1];
    }
    var diff = maxent[0] - scale[idx][0];
    if (diff == 0) {
        diff = 1;
    }
    var degree = cur - scale[idx][0]; // offset into the range of temps
    var factor = degree / diff; // normalize to 0 to 1
    if (factor < 0) { factor = 0; }
    if (factor > 1) { factor = 1; }
    var maxhue = scale[idx][1];
    var minhue = maxent[1];
    var deltahue = maxhue - minhue;
    factor = ((1 - factor) * deltahue) + minhue; // 0 is red, 240 is blue, follow the hue wheel
    return "hsl(" + factor + ", 100%, 50%)";
}

function placeNode(currloc, name) {
    var currack;
    var currow;
    if ('row' in currloc) {
        currow = currloc['row'];
    } else { // probably single row
        currow = ' ';
    }
    if (!rows.hasOwnProperty(currow)) {
        rows[currow] = {};
    }
    if ('rack' in currloc) {
        currack = currloc['rack'];
    } else { //probably single rack
        currack = ' ';
    }
    if (!rows[currow].hasOwnProperty(currack)) {
        rows[currow][currack] = {};
    }
    rows[currow][currack][currloc['u']] = name;
}

function orglocations() {
    rows = {};
    enclosures = {};
    for (var node in nodelocations) {
        var currloc = nodelocations[node];
        var currow;
        if ('bay' in currloc) {
            var curenc = currloc['enclosure'];
            if (!enclosures.hasOwnProperty(curenc)) {
                enclosures[curenc] = {}
            }
            enclosures[curenc][currloc['bay']] = node;
            if ('u' in currloc) { // enclosure locality on node in effect...
                // place the enclosure based on the location properties
                // of the node
                placeNode(currloc, curenc);
            }
        } else {
            placeNode(currloc, node);
        }
    }
}

function addNodeLocation(data) {
    if (!Array.isArray(data.databynode)) {
        data.databynode = [data.databynode];
    }
    nodelocations = {};
    for (var i = 0; i < data.databynode.length; i ++) {
        var nodedata = data.databynode[i];
        for (var node in nodedata) {
            if (!nodelocations.hasOwnProperty(node)) {
                nodelocations[node] = {};
            }
            var currdata = nodedata[node];
            if ('enclosure.bay' in currdata) {
                nodelocations[node]['bay'] = currdata['enclosure.bay']['value']
            }
            if ('enclosure.manager' in currdata) {
                nodelocations[node]['enclosure'] = currdata['enclosure.manager']['value']
            }
            if ('location.rack' in currdata) {
                nodelocations[node]['rack'] = currdata['location.rack']['value']
            }
            if ('location.u' in currdata) {
                nodelocations[node]['u'] = currdata['location.u']['value']
            }
            if ('location.row' in currdata) {
                nodelocations[node]['row'] = currdata['location.row']['value']
            }
        }
    }
    orglocations();
    drawNodes();
    colorNodes();
}
var mainpane;
var active;
var fetchdata;

var currdivs = [];

function colorDiv(data) {
    var sensoridx = this[1];
    var thenoverlay = this[0];
    if (thenoverlay !== curroverlay)  { return; } // Cease drawing stale responses that would mess up
    if (!active) { return; }
    for (var nname in data['databynode']) {
        if (!data['databynode'][nname].hasOwnProperty('sensors')) {
            continue;
        }
        var hotness = data['databynode'][nname].sensors[0].value;
        if (hotness == null) { continue; }
        var units = data['databynode'][nname].sensors[0].units;
        if (!nodedivs.hasOwnProperty(nname)) {
            continue;
        }
        if (minvalue === null || minvalue > hotness) { minvalue = hotness; }
        if (maxvalue === null || maxvalue < hotness) { maxvalue = hotness; }
        currdivs.push(nodedivs[nname]);
        nodedivs[nname].getElementsByTagName('span')[1].textContent = hotness + units;
        nodedivs[nname].hotness = hotness;
        if (curroverlay.scale === null) {
             // auto-scale should recolor all currdivs
             currdivs.forEach(function(thediv) {
                  thediv.style.background = generateColor(curroverlay.scale, thediv.hotness);
             });
        } else {
             nodedivs[nname].style.background = generateColor(curroverlay.scale, hotness);
        }
    }
    if ('_requestdone' in data) {
        currdivs = [];
        fetchdata = setTimeout(colorNodes, 1500);
    }
}

function colorNodes() {
    fetchdata = null;
    if (!active) {
        return;
    }
    getRequest('/confluent-api/noderange/' + noderange + '/sensors/hardware/all/' + curroverlay['sensors'][0], colorDiv.bind([curroverlay, 0]));
}

function addNodeToEnclosure(nodename, container, bay) {
    var topkey = (1 - (Math.ceil(bay/2) / 6)) * 100;
    var nodediv = document.createElement('div');
    var nodenamespan = document.createElement('span');
    nodenamespan.appendChild(document.createTextNode(nodename));
    nodediv.appendChild(nodenamespan);
    nodediv.style.transition = '1.5s';
    var dataspan = document.createElement('span');
    dataspan.style.float = 'right';
    nodediv.appendChild(dataspan);
    nodedivs[nodename] = nodediv;
    nodediv.style.top = topkey + "%";
    nodediv.style.height = (100/6) + '%';
    nodediv.style.position = 'absolute';
    nodediv.style.width = '49.5%';
    nodediv.style.margin = '0';
    nodediv.style.borderWidth = "1px";
    nodediv.style.borderStyle = "solid";
    if (bay % 2) {
        nodediv.style.float = 'left';
    } else {
        nodediv.style.left = '50%';
    }
    container.appendChild(nodediv);
}

function createEnclosure(myrack, u, name) {
    //TODO: non-nextscale geometry
    var theheight = 600/racku;
    var thetop =  99.5 - (((parseInt(u) - 1) * 100/racku) + theheight);
    var encdiv = document.createElement('div');
    encdiv.style.height = theheight + "%";
    encdiv.style.top = thetop + "%";
    encdiv.style.width = "100%";
    encdiv.style.position = "absolute";
    encdiv.style.borderStyle = "solid";
    encdiv.style.borderWidth = "1px";
    encdiv.style.margin = '0';
    for (var bay in enclosures[name]) {
        var nodename = enclosures[name][bay];
        addNodeToEnclosure(nodename, encdiv, bay);

    }
    myrack.appendChild(encdiv);
}

function drawNodes() {
    for (var row in rows) {
        mainpane.appendChild(document.createElement('hr'));
        if (row != " ") {
            mainpane.appendChild(document.createTextNode('Row {}'.l(row)));
            mainpane.appendChild(document.createElement('br'));
        }
        for (var rack in rows[row]) {
            var rackdata = rows[row][rack];
            var myrack = createRack(rack);
            for (var curu in rackdata) {
                if (curu !== undefined) {
                    var curnode = rackdata[curu];
                    if (curnode in enclosures) {
                        createEnclosure(myrack, curu, curnode);
                    }
                }
            }
        }
    }
}

function getNodeLocations() {
  minvalue = null;
  maxvalue = null;
  getRequest('/confluent-api/noderange/' + noderange + '/attributes/current',
             addNodeLocation, donull, undefined, undefined, true);
}
    
    function closerackview() {
        active = false;
        if (nrchanger) {
            clearTimeout(nrchanger);
            nrchanger = null;
        }
        if (fetchdata) {
            clearTimeout(fetchdata);
            fetchdata = null;
        }
    }
    function rackview() {
        clear_main();
        active = true;
        window.main = rackview;
        window.closemain = closerackview;
        var truemainpane = document.getElementById('main');
        truemainpane.appendChild(maketoolbar());
        mainpane = document.createElement('div');
        truemainpane.appendChild(mainpane);
        nodedivs = {};
        if (!curroverlay) {
            curroverlay = overlaysensors['Ambient Temperature'];
        }
        overlayselect.header.labeltext.textContent = 'Overlay: {} '.l(l('Ambient Temperature'));
        getNodeLocations();

    };

    window.rackview = rackview;
})();
